import qree;
import pytest;

# Run tests with:
#       pytest tests.py -vv

def test_identity ():
    assert qree.renderStr("plain") == "plain";
    assert qree.renderStr('foo\nbar\nbaz') == 'foo\nbar\nbaz';

def test_basic_substitution ():
    tpl = "Hello, {{: data + '!' :}}";
    assert qree.renderStr(tpl, "World") == 'Hello, World!';
    assert qree.renderStr(tpl, "> World") == 'Hello, &gt; World!';
    tpl2 = "Hello, {{= data + '!' =}}";
    assert qree.renderStr(tpl2, "World") == 'Hello, World!';
    assert qree.renderStr(tpl2, "> World") == 'Hello, > World!';

def test_nonPyQuotesOk ():
    "Test that single-quotes _outside_ py-code are OK.";
    qree.renderStr("Trip ''' quotes") == "Trip ''' quotes";
    qree.renderStr("'{{: data :}}'", None) == "'None'";

def test_pyQuotesOk ():
    "Test that single-quotes _within_ py-code are OK.";
    qree.renderStr("{{: data['foo'] :}}", {"foo": "bar"}) == "bar";
    qree.renderStr("{{: data['''foo'''] :}}", {"foo": "bar"}) == "bar";

def test_substitution_tag_error ():
    badTplList =[
        "Hello, {{: data + '!' ]]",             # Unclosed
        "Hello, {{= {{: data + '!' :}} =}}",    # Nesting banned
        "Hello, {{: {{= data + '!' =}} :}}",
        """
        @= import qree;
        {{: qree.renderStr('{{= data.title() =}}', data=data) :}}
        """,
    ];
    for badTpl in badTplList:
        with pytest.raises(SyntaxError):
            qree.renderStr(badTpl, "World");

def test_indent_tag_error ():
    badTplList = [
        # 1, 2 :
        "@{", "@}",
        # --------------------------
        # 3:
        """
        @= if True:
        @{
            Unclosed indent.
        #
        """,
        # --------------------------
        # 4:
        """
        @= if True:
        #
            Unopened indent.
        @}
        """,
        # --------------------------
        # 5:
        """
        @= if True:
        @{
            if True:
            @{
                Unclosed inner indent.
            #
        @}
        """,
        # --------------------------
        # 6:
        """
        @= if True:
        @{ Non-comment line along w/ open-indent.
        @}
        """,
        # --------------------------
        # 7:
        """
        @= if True:
        @{
        @} Non-comment line along w/ close-indent.
        """,
    ];
    assert len(badTplList) == 7;
    for badTpl in badTplList:
        with pytest.raises(IndentationError):
            print(qree.renderStr(badTpl, data=None));

def test_basic_indent ():
    tpl = """
    @= for n in range(1, data+1):
    @{
        {{: n :}} is {{: 'EVEN' if n % 2 == 0 else 'ODD' :}}
    @}
    """;
    expected = """
        1 is ODD
        2 is EVEN
        3 is ODD
        4 is EVEN
    """;
    assert qree.renderStr(tpl, data=4) == expected;

def test_nested_indent ():
    tpl = """
    @= for n in range(1, data+1):
    @{
        @= if n % 15 == 0:
        @{  # Comment w/ indent-open.
            FizzBuzz!
        @}
        @= elif n % 3 == 0:
        @{
            Fizz!
        @}  # Comment w/ indent-close.
        @= elif n % 5 == 0:
        @{
            Buzz!
        @}
        @= else:
        @{
            {{: n :}}
        @}
    @}
    """;
    expected = """
            1
            2
            Fizz!
            4
            Buzz!
            Fizz!
            7
            8
            Fizz!
            Buzz!
            11
            Fizz!
            13
            14
            FizzBuzz!
    """;
    assert qree.renderStr(tpl, data=15) == expected;
    

def test_renderPath_basic_plus_with_view ():
    footerPath = "./test-views/footer.html";
    expectedFooter = "\n".join([
        '<footer class="footer">',
        '    <a href="/linkA">Link A</a>',
        '    <a href="/linkB">Link B</a>',
        '    <a href="/linkC">Link C</a>',
        '</footer>',
        '', # <-- Qree seems to like adding trailing newlines.
    ]);
    print(repr(expectedFooter));
    print(repr(qree.renderPath(footerPath)));
    assert qree.renderPath(footerPath) == expectedFooter;
    
    @qree.view(footerPath)
    def serveFooter ():
        return None;
    assert serveFooter() == expectedFooter;


def test_renderPath_nested_plus_with_view ():
    homepagePath = "./test-views/homepage.html";
    data = {
        "title": "... TITLE 2 ...",
        "body": "... BODY 2 ...",
    }
    expectedHomepage = "\n".join([
        '<doctype html>',
        '<html>',
        '<head><title>... TITLE 2 ...</title></head>',
        '<body>',
        '<header class="header">',
        '    <a href="/link1">Link 1</a>',
        '    <a href="/link2">Link 2</a>',
        '    <a href="/link3">Link 3</a>',
        '</header>',
        '',                             # <-- \n via nested call.
        '<h2>... TITLE 2 ...</h2>',
        '<pre>... BODY 2 ...</pre>',
        '<footer class="footer">',
        '    <a href="/linkA">Link A</a>',
        '    <a href="/linkB">Link B</a>',
        '    <a href="/linkC">Link C</a>',
        '</footer>',
        '',                             # <-- \n via nested call.
        '</body>',
        '<html>',
        '', # <-- Qree seems to like adding trailing newlines.
    ]);
    assert qree.renderPath(homepagePath, data) == expectedHomepage;
    
    @qree.view(homepagePath)
    def serveHomepage ():
        return data;
    assert serveHomepage() == expectedHomepage;
    
    print(expectedHomepage);
test_renderPath_nested_plus_with_view();

# End ######################################################



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='qree',
      version='0.0.3',
      description='Qree: Tiny but mighty Python templating.',
      author='Polydojo, Inc.',
      author_email='8079593+sumukhbarve@users.noreply.github.com',
      url='https://github.com/polydojo/qree',
      py_modules=['qree'],
     )



"""
Qree: Tiny but mighty Python templating.

Copyright (c) 2020 Polydojo, Inc.

SOFTWARE LICENSING
------------------
The software is released "AS IS" under the MIT License,
WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED. Kindly
see LICENSE.txt for more details.

NO TRADEMARK RIGHTS
-------------------
The above software licensing terms DO NOT grant any right in the
trademarks, service marks, brand names or logos of Polydojo, Inc.
""";

import functools;

__version__ = "0.0.3";  # Req'd by flit.
__DEFAULT_TAG_MAP__ = { "@=": "@=",  "@{": "@{",  "@}": "@}",
    "{{:": "{{:",  ":}}": ":}}",  "{{=": "{{=",  "=}}": "=}}",
};

escHtml = lambda s: (str(s).replace("&", "&amp;").replace("<", "&lt;")
    .replace(">", "&gt;").replace('"', "&quot;").replace("'", "&#x27;")          # TODO: Consider: .replace("`", "&#x60;")
);

def dictDefaults (dicty, defaults):
    "Fills-in missing keys in `dicty` with those from `defaults`.";
    for k in defaults:
        if k not in dicty:
            dicty[k] = defaults[k];
    return dicty;

def findFirstMatch(haystack, needleList, fromIndex=0):
    "Finds needle in `haystack from needlList w/ lowest index.";
    foundNeedle = None;
    minIndex = None;
    for needle in needleList:
        index = haystack.find(needle, fromIndex);
        if index == -1:
            pass;
        elif (minIndex is None) or (index < minIndex):
            foundNeedle = needle;
            minIndex = index;
    return (foundNeedle, minIndex);

def validateSubstitutionTagPair (opTag, clTag, tagMap):
    "Helps escapeNonPyQuotes().";
    assert opTag in [tagMap["{{="], tagMap["{{:"]];
    expectedClTag = tagMap["=}}" if opTag == tagMap["{{="] else ":}}"];
    if clTag != expectedClTag:
        raise SyntaxError("Tag-mismatch. Expected %r, not %r" % (
            expectedClTag, clTag,
        ));
    # otherwise ...
    return True;

def escapeNonPyQuotes (line, tagMap):
    tags = list(map(tagMap.get, "{{= =}} {{: :}}".split()));
    firstTag, firstIndex = findFirstMatch(line, tags);
    if not firstTag:
        return line.replace("'", r"\'");
    # otherwise ...
    nextTag, nextIndex = findFirstMatch(line, tags, firstIndex+1);
    assert validateSubstitutionTagPair(firstTag, nextTag, tagMap);
    beyondNextTagEndIndex = nextIndex + len(nextTag);
    return (
        line[ : firstIndex].replace("'", r"\'") +
        line[firstIndex: beyondNextTagEndIndex] +
        escapeNonPyQuotes(line[beyondNextTagEndIndex : ], tagMap) #+
    );

def validateStandaloneIndentLine (line, tag):
    "Ensures indent-line only has indent-tag, excl. comment.";
    if line.split("#")[0].strip() != tag:
        raise IndentationError("Invalid de/indent line: %r" % line);
    return True;

def quoteReplace (tplStr, variable="data", tagMap=None):
    "Returns as string, the function-equivalent of `tplStr`.";
    tagMap = dictDefaults(tagMap or {}, __DEFAULT_TAG_MAP__);
    fnStr = "def templateFn (%s):\n" % variable;
    innerIndentDepth = 0;   # Depth due to @{ and @} only.
    indentify = lambda: " " * ((1 + innerIndentDepth) * 4);
    fnStr += indentify() + "from qree import escHtml as __qreeEsc;\n";
    fnStr += indentify() + "output = '';\n";
    for line in tplStr.splitlines(True):
        # Param `keepends=True` ^^^^. (Not kwarg for py2.)
        lx = line.lstrip();
        if lx.startswith(tagMap["@="]):
            pyCode = lx[len(tagMap["@="]) : ].strip();
            fnStr += indentify() + pyCode + "\n";
        elif lx.startswith(tagMap["@{"]):
            assert validateStandaloneIndentLine(lx, tagMap["@{"])
            innerIndentDepth += 1;
        elif lx.startswith(tagMap["@}"]):
            assert validateStandaloneIndentLine(lx, tagMap["@}"])
            innerIndentDepth -= 1;
        else:
            
            fnStr += indentify() + "output += " + "'''" +  (
                escapeNonPyQuotes(line, tagMap) # <- Err thrower
                    .replace(tagMap["{{="],  "''' + str(")
                    .replace(tagMap["=}}"],  ") + '''")
                    .replace(tagMap["{{:"],  "''' + __qreeEsc(")
                    .replace(tagMap[":}}"],  ") + '''")
            ) + "''';\n";
    fnStr += indentify() + "return output;\n";
    if innerIndentDepth != 0:
        raise IndentationError("Tag-mismatch for tags " +
            ("%r and %r." % (tagMap["@{"], tagMap["@}"])) #+
        );
    return fnStr;

def execEval (fnStr):
    "Converts `fnStr` to a callable function and returns it.";
    exec(fnStr); # Via exec(), 'templateFn' has been defined.
    return eval("templateFn");

def renderStr (tplStr, data=None, variable="data", tagMap=None):
    "Render template `tplStr` using `data`.";
    fnStr = quoteReplace(tplStr, variable, tagMap);
    fn = execEval(fnStr);
    return fn(data);

def renderPath (tplPath, data=None, variable="data", tagMap=None):
    "Render template at path `tplPath` using `data`.";
    with open(tplPath, "r") as f:
        return renderStr(f.read(), data, variable, tagMap);

def view (tplPath, variable="data", tagMap=None):
    "Returns a decorator for binding function to template at `tplPath`.";
    def decorator (fn):
        @functools.wraps(fn)
        def wrapper (*a, **ka):
            return renderPath(tplPath, fn(*a, **ka), variable, tagMap);
        return wrapper;
    return decorator;

# TODO/Consider:
#checkNonPathy = lambda s: "\n" in s or "{" in s or "@" in s;
#def render (s, data=None, variable="data", tagMap=None):
#    "Wrapper that auto-picks renderPath() or renderStr()";
#    renderFn = renderStr if checkNonPathy(s) else renderPath;
#    return renderFn(s, data, variable, tagMap);

# End ######################################################
